// apt.cc
//
//  Copyright 2000 Daniel Burrows
//
//  The main file for python-apt.

#include <python/Python.h>

#include <apt-pkg/init.h>
#include <apt-pkg/strutl.h>
#include <apt-pkg/configuration.h>
#include <apt-pkg/progress.h>
#include <apt-pkg/cachefile.h>
#include <apt-pkg/depcache.h>
#include <apt-pkg/error.h>

static PyObject *AptPyExc_AptError;

static PyObject *AptPyStringToBool(PyObject *self, PyObject *args)
{
  char *str;

  if(!PyArg_ParseTuple(args, "s", &str))
    return NULL;

  // FIXME: throw an exception if StringToBool returns -1 (error)?
  return Py_BuildValue("i", StringToBool(str));
}

static PyObject *AptPyStringToNum(PyObject *self, PyObject *args)
{
  char *str;
  unsigned long i;

  if(!PyArg_ParseTuple(args, "s", &str))
    return NULL;

  if(StrToNum(str, i, strlen(str)))
    return PyInt_FromLong(i);
  else
    return PyInt_FromLong(0);
  // Should I throw an exception?
}

/*****************************************************************************
 *  Configuration item class wrapper
 *****************************************************************************/

struct AptPyConfigItem
// Used to access the configuration tree structure.
{
  PyObject_HEAD;
  const Configuration::Item *item;
};

static PyObject *AptPyConfigItem_new(const Configuration::Item *item);

static void AptPyConfigItem_dealloc(AptPyConfigItem *self)
{
  PyMem_DEL(self);
}

static PyObject *AptPyConfigItem_getattr(AptPyConfigItem *self, char *name)
{
  if(!strcmp(name, "value"))
    return PyString_FromString(self->item->Value.c_str());
  else if(!strcmp(name, "tag"))
    return PyString_FromString(self->item->Tag.c_str());
  else if(!strcmp(name, "parent"))
    return self->item->Parent?AptPyConfigItem_new(self->item->Parent):Py_None;
  else if(!strcmp(name, "child"))
    return self->item->Child?AptPyConfigItem_new(self->item->Child):Py_None;
  else if(!strcmp(name, "next"))
    return self->item->Next?AptPyConfigItem_new(self->item->Next):Py_None;
  else if(!strcmp(name, "fulltag"))
    return PyString_FromString(self->item->FullTag().c_str());

  PyErr_SetString(PyExc_AttributeError, name);
  return NULL;
}

static int AptPyConfigItem_length(AptPyConfigItem *self)
{
  int len=0;
  for(Configuration::Item *i=self->item->Child; i; i=i->Next)
    len++;

  return len;
}

static PyObject *AptPyConfigItem_item(AptPyConfigItem *self, int i)
{
  Configuration::Item *itm;

  if(i<0)
    {
      PyErr_SetString(PyExc_IndexError, "index out of range");
      return NULL;
    }

  for(itm=self->item->Child; itm && i>0; itm=itm->Next, i--)
    ;

  if(!itm)
    {
      PyErr_SetString(PyExc_IndexError, "index out of range");
      return NULL;
    }

  return AptPyConfigItem_new(itm);
}

static PySequenceMethods AptPyConfigItem_as_sequence=
{
  (inquiry) AptPyConfigItem_length, // sq_length
  NULL, // sq_concat
  NULL, // sq_repeat
  (intargfunc) AptPyConfigItem_item, // sq_item
  NULL, // sq_slice
  NULL,
  NULL // Assignment stuff
};

static PyTypeObject AptPyConfigItem_type=
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "AptConfigItem",
  sizeof(AptPyConfigItem),
  0,
  (destructor) AptPyConfigItem_dealloc,
  0,
  (getattrfunc) AptPyConfigItem_getattr,
  0,
  0,
  0,
  0,
  &AptPyConfigItem_as_sequence,
  0
};

static PyObject *AptPyConfigItem_new(const Configuration::Item *item)
{
  AptPyConfigItem *self=(AptPyConfigItem *) PyObject_NEW(AptPyConfigItem, &AptPyConfigItem_type);

  if(self==NULL)
    return NULL;
  self->item=item;
  return (PyObject *) self;
}

/*****************************************************************************
 *  Configuration class wrapper
 *****************************************************************************/

struct AptPyConfig
// Wraps the Configuration class..
{
  PyObject_HEAD;
  Configuration *config;
};

static void AptPyConfig_dealloc(AptPyConfig *self)
{
  delete self->config;
  PyMem_DEL(self);
}

static PyObject *AptPyConfig_subscript(AptPyConfig *self, PyObject *key)
{
  if(PyString_Check(key))
    return PyString_FromString(self->config->Find(PyString_AsString(key)).c_str());
  else
    {
      PyErr_SetObject(PyExc_TypeError, key);
      return NULL;
    }
}

static PyMappingMethods AptPyConfig_as_mapping=
{
  0,
  (binaryfunc) AptPyConfig_subscript,
  0
};

static PyObject *AptPyConfig_find(AptPyConfig *self, PyObject *args)
  // Wraps Configuration::Find
{
  char *loc,*def;

  if(!PyArg_ParseTuple(args, "ss", &loc, &def))
    return NULL;

  return PyString_FromString(self->config->Find(loc, def).c_str());
}

static PyObject *AptPyConfig_tree(AptPyConfig *self, PyObject *args)
  // Returns a configuration tree
{
  char *loc;

  if(!PyArg_ParseTuple(args, "s", &loc))
    return NULL;

  const Configuration::Item *rval=self->config->Tree(loc);

  if(!rval)
    {
      PyErr_SetString(PyExc_KeyError, loc);
      return NULL;
    }
  else
    return AptPyConfigItem_new(self->config->Tree(loc));
}

static PyObject *AptPyConfig_set(AptPyConfig *self, PyObject *args)
  // Wraps Configuration::Find
{
  char *loc,*def;

  if(!PyArg_ParseTuple(args, "ss", &loc, &def))
    return NULL;

  self->config->Set(loc, def);

  return Py_None;
}

static PyMethodDef AptPyConfig_methods[]=
{
  {"find", (PyCFunction) AptPyConfig_find, METH_VARARGS},
  {"set", (PyCFunction) AptPyConfig_set, METH_VARARGS},
  {"tree", (PyCFunction) AptPyConfig_tree, METH_VARARGS},
  {NULL, NULL}
};

static PyObject *AptPyConfig_getattr(AptPyConfig *self, char *name)
{
  PyObject *rval;

  if((rval=Py_FindMethod(AptPyConfig_methods, (PyObject *) self, name)))
    return rval;
  else
    return rval;
  // FIXME: automagic config stuff goes here.
}

static PyTypeObject AptPyConfig_type=
{
  PyObject_HEAD_INIT(&PyType_Type)
  0, // Say what?  What's this for?
  "AptConfiguration",			// tp_name
  sizeof(AptPyConfig),			// tp_basicsize
  0,					// tp_itemsize (undocumented)
  (destructor) AptPyConfig_dealloc,	// tp_dealloc
  0,					// tp_print
  (getattrfunc) AptPyConfig_getattr,	// tp_getattr
  0,					// tp_setattr
  0,					// tp_compare
  0,					// tp_repr
  0,					// tp_as_number
  0,					// tp_as_sequence
  &AptPyConfig_as_mapping,		// tp_hash??..no, tp_as_mapping
  0,					// tp_call
  0,					// tp_str
  0,					// tp_getattro
  0,					// tp_setattro
  0,					//  ???????
  0,					// tp_as_buffer
  Py_TPFLAGS_DEFAULT,			// tp_flags
  "The APT configuration class"
};

static PyObject *AptPyConfig_new(Configuration *config)
{
  AptPyConfig *self=(AptPyConfig *)PyObject_NEW(AptPyConfig, &AptPyConfig_type);

  if(self==NULL)
    return NULL;
  self->config=config;
  return (PyObject *) self;
}

/*****************************************************************************
 *  Error class wrapper
 *****************************************************************************/

struct AptPyError
// Wraps the Error class
{
  PyObject_HEAD;
  GlobalError *error;
};

static void AptPyError_dealloc(AptPyError *self)
{
  delete self->error;
  PyMem_DEL(self);
}

static PyObject *AptPyError_pop(AptPyError *self)
  // Returns a tuple (msg, error) consisting of the message and a flag
  // which is true if the message is an error.
{
  if(self->error->empty())
    {
      PyErr_SetString(PyExc_IndexError, "Pop from empty message queue");
      return NULL;
    }
  else
    {
      string msg;
      bool is_error=self->error->PopMessage(msg);
      return Py_BuildValue("si", msg.c_str(), is_error?1:0);
    }
}

static PyObject *AptPyError_discard(AptPyError *self)
{
  self->error->Discard();
  return Py_None;
}

static PyObject *AptPyError_dumperrors(AptPyError *self)
{
  self->error->DumpErrors();
  return Py_None;
}

static PyMethodDef AptPyError_methods[]=
{
  {"pop", (PyCFunction) AptPyError_pop, METH_VARARGS},
  {"discard", (PyCFunction) AptPyError_discard, METH_VARARGS},
  {"dumperrors", (PyCFunction) AptPyError_dumperrors, METH_VARARGS}
};

static PyObject *AptPyError_getattr(AptPyError *self, char *name)
{
  if(!strcmp(name, "pendingerror"))
    return PyInt_FromLong(self->error->PendingError());
  else if(!strcmp(name, "empty"))
    return PyInt_FromLong(self->error->empty());
  else
    return Py_FindMethod(AptPyError_methods, (PyObject *) self, name);
}

static PyTypeObject AptPyError_type=
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "AptGlobalError",			// tp_name
  sizeof(AptPyError),			// tp_basicsize
  0,					// tp_itemsize (undocumented)
  (destructor) AptPyError_dealloc,	// tp_dealloc
  0,					// tp_print
  (getattrfunc) AptPyError_getattr,	// tp_getattr
  0,					// tp_setattr
  0,					// tp_compare
  0,					// tp_repr
  0,					// tp_as_number
  0,					// tp_as_mapping
  0,					// tp_hash
  0,					// tp_call
  0,					// tp_str
  0,					// tp_getattro
  0,					// tp_setattro
  0,					//  ???????
  0,					// tp_as_buffer
  Py_TPFLAGS_DEFAULT,			// tp_flags
  "A class to store a global stack of errors"
};

static PyObject *AptPyError_new(GlobalError *error)
{
  AptPyError *self=(AptPyError *) PyObject_NEW(AptPyError, &AptPyError_type);

  if(!self)
    return NULL;

  self->error=error;
  return (PyObject *) self;
}

/*****************************************************************************
 *  Progress-bar glue/wrapper
 *****************************************************************************/

//  Two things are needed here: first, a way to represent 'real' progress
// bars from the Python side, and second, a way for Python objects to act
// as progress bars.  In terms of the first, it's tempting to simply wrap
// the progress bar in a Python C object, then use it the same way a "pure"
// Python progress bar is used; however, this won't work -- the internal status
// information of the progress bar won't be properly updated.
//  The implementation chosen is not what I'd most prefer, but it works; the
// idea is that we wrap the C++ progress-bar objects in a Python object, so
// that they can be used as progress bars, and that it's also possible to use a
// Python object as a progress bar -- various methods and attributes are used
// to set up the interface.  In other words, C++ <-> Python routines that can
// accept progress bars need to handle both cases; a handy routine is provided
// to do this (Pure Python routines or pure C++ routines don't need to handle
// both cases, as the differences are only visible in here)

// Wraps a C++ progress bar in a Python object
struct AptPyOpProgress
{
  PyObject_HEAD;
  OpProgress *progress;
};

static PyObject *AptPyOpProgress_new(OpProgress *progress);

static void AptPyOpProgress_dealloc(AptPyOpProgress *self)
{
  delete self->progress;
  // Hmmmmm, this could be problematic..what if C++ code 'owns' this progress
  // bar?
  PyMem_DEL(self);
}

static PyObject *AptPyOpProgress_progress(AptPyOpProgress *self, PyObject *args)
{
  long current;

  if(!PyArg_ParseTuple(args, "l", &current))
    return NULL;

  self->progress->Progress(current);
  return Py_None;
}

static PyObject *AptPyOpProgress_subprogress(AptPyOpProgress *self, PyObject *args)
{
  long subtotal;
  char *op=NULL;

  if(!PyArg_ParseTuple(args, "l|s", &subtotal, &op))
    return NULL;

  if(!op)
    self->progress->SubProgress(subtotal);
  else
    self->progress->SubProgress(subtotal, op);

  return Py_None;
}

static PyObject *AptPyOpProgress_overallprogress(AptPyOpProgress *self, PyObject *args)
{
  long current;
  long total;
  long size;
  char *op;

  if(!PyArg_ParseTuple(args, "llls", &current, &total, &size, &op))
    return NULL;

  self->progress->OverallProgress(current, total, size, op);

  return Py_None;
}

static PyObject *AptPyOpProgress_done(AptPyOpProgress *self, PyObject *args)
{
  if(!PyArg_ParseTuple(args, ""))
    return NULL;

  self->progress->Done();

  return Py_None;
}

static PyMethodDef AptPyOpProgress_methods[]=
{
  {"progress", (PyCFunction) AptPyOpProgress_progress, METH_VARARGS},
  {"subprogress", (PyCFunction) AptPyOpProgress_subprogress, METH_VARARGS},
  {"overallprogress", (PyCFunction) AptPyOpProgress_overallprogress, METH_VARARGS},
  {"done", (PyCFunction) AptPyOpProgress_done, METH_VARARGS}
};

static PyObject *AptPyOpProgress_getattr(AptPyOpProgress *self, char *name)
{
  return Py_FindMethod(AptPyOpProgress_methods, (PyObject *) self, name);
}

static PyTypeObject AptPyOpProgress_type=
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "AptOpProgress",
  sizeof(AptPyOpProgress),
  0,
  (destructor) AptPyOpProgress_dealloc,
  0,
  (getattrfunc) AptPyOpProgress_getattr,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  Py_TPFLAGS_DEFAULT,
  "Generic progress display object for APT"
};

static PyObject *AptPyOpProgress_new(OpProgress *progress)
{
  AptPyOpProgress *self=(AptPyOpProgress *) PyObject_NEW(AptPyOpProgress, &AptPyOpProgress_type);

  if(!self)
    return NULL;

  self->progress=progress;
  return (PyObject *) self;
}

static PyObject *AptPyOpProgress_generic(PyObject *self, PyObject *args)
{
  if(!PyArg_ParseTuple(args, ""))
    return NULL;

  return AptPyOpProgress_new(new OpProgress);
}

static PyObject *AptPyOpProgress_text(PyObject *self, PyObject *args)
{
  if(!PyArg_ParseTuple(args, ""))
    return NULL;

  return AptPyOpProgress_new(new OpTextProgress);
}

/*****************************************************************************
 *  Class allowing a Python object to act as a progress display
 *
 *  The object must provide methods "done" and "update", which are called as
 * with C++ OpProgress subclasses.  The attributes 'op', 'subop', 'percent',
 * and 'majorchange' are set by the C++ object before calling the routines;
 * 'checkchange' can, unfortunately, not be directly accessed from Python;
 * I compromised by calling it for each of these operations and then setting
 * the 'ischange' attribute.
 *****************************************************************************/

class OpPyProgress:public OpProgress
{
  PyObject *py_val;
  // The Python object being used.

  void UpdateAttrs();

protected:
  virtual void Update() {UpdateAttrs(); PyObject_CallMethod(py_val, "update", "");}
public:
  OpPyProgress(PyObject *_py_val):py_val(_py_val)
  {
    Py_INCREF(py_val);
  }
  virtual ~OpPyProgress() {Py_DECREF(py_val);}

  virtual void Done() {UpdateAttrs(); PyObject_CallMethod(py_val, "done", "");}
};

void OpPyProgress::UpdateAttrs()
{
  // FIXME: this is somewhat inefficient..maybe I should check whether the
  //       values have changed? (so as to avoid allocating/deleting tons of
  //       memory each time this is called)
  //
  //  ..hmmm..SetAttrString doesn't play with refcounts, does it?
  PyObject_SetAttrString(py_val, "op", PyString_FromString(Op.c_str()));
  PyObject_SetAttrString(py_val, "subop", PyString_FromString(Op.c_str()));
  PyObject_SetAttrString(py_val, "percent", PyFloat_FromDouble(Percent));
  PyObject_SetAttrString(py_val, "majorchange", PyInt_FromLong(MajorChange));

  PyObject_SetAttrString(py_val, "ischange", PyInt_FromLong(CheckChange()));
}

bool is_progbar(PyObject *obj)
{
  PyObject *update=PyObject_GetAttrString(obj, "update");
  if(!update || !PyCallable_Check(update))
    return false;

  PyObject *done=PyObject_GetAttrString(obj, "done");
  if(!done || !PyCallable_Check(done))
    return false;

  return true;
}

PyObject *AptPyOpProgress_fromobj(PyObject *self, PyObject *args)
{
  PyObject *obj;

  if(!PyArg_ParseTuple(args, "O", &obj))
    return NULL;

  if(!is_progbar(obj))
    return NULL;

  return AptPyOpProgress_new(new OpPyProgress(obj));
}

/*****************************************************************************
 *  Forward declarations of iterator-creation routines
 *****************************************************************************/

static PyObject *AptPyPkgIterator_new(pkgCache::PkgIterator pkg);
static PyObject *AptPyVerIterator_new(pkgCache::VerIterator ver);

/*****************************************************************************
 *  PkgIterator wrapper
 *****************************************************************************/

struct AptPyPkgIterator
{
  PyObject_HEAD;
  pkgCache::PkgIterator pkg;
};

static PyObject *AptPyPkgIterator_dealloc(AptPyPkgIterator *self)
{
  PyMem_DEL(self);
}

static int AptPyPkgIterator_compare(AptPyPkgIterator *a, AptPyPkgIterator *b)
{
  return a->pkg!=b->pkg;
}

static PyObject *AptPyPkgIterator_repr(AptPyPkgIterator *self)
  // FIXME: should this be 'str' instead of 'repr'?
{
  return PyString_FromString((string("<APT package iterator: ")+(self->pkg.end()?"END":self->pkg.Name())+">").c_str());
}

static PyObject *AptPyPkgIterator_getattr(AptPyPkgIterator *self, char *name)
{
  if(!strcmp(name, "end"))
    return PyInt_FromLong(self->pkg.end());

  if(self->pkg.end())
    {
      PyErr_SetString(PyExc_IndexError, "Attempt to index off the end of a package cache");
      return NULL;
    }

  if(!strcmp(name, "next"))
    // You can't advance the Python object like the C++ iterator because (a)
    // Python doesn't handle things that way and (b) a=b; a.gonext() wouldn't
    // do what you expect (overriding the assignment operator might help
    // in that case..)
    {
      pkgCache::PkgIterator next=self->pkg;
      next++;
      return AptPyPkgIterator_new(next);
    }
  else if(!strcmp(name, "name"))
    return PyString_FromString(self->pkg.Name());
  else if(!strcmp(name, "section"))
    return PyString_FromString(self->pkg.Section());
  else if(!strcmp(name, "targetdist"))
    return PyString_FromString(self->pkg.TargetDist());
  else if(!strcmp(name, "state"))
    return PyInt_FromLong(self->pkg.State());
  else if(!strcmp(name, "selectedstate"))
    return PyInt_FromLong(self->pkg->SelectedState);
  else if(!strcmp(name, "inststate"))
    return PyInt_FromLong(self->pkg->InstState);
  else if(!strcmp(name, "currentstate"))
    return PyInt_FromLong(self->pkg->CurrentState);
  else if(!strcmp(name, "id"))
    return PyInt_FromLong(self->pkg->ID);
  else if(!strcmp(name, "flags"))
    return PyInt_FromLong(self->pkg->Flags);
  else if(!strcmp(name, "versionlist"))
    return AptPyVerIterator_new(self->pkg.VersionList());
  else if(!strcmp(name, "targetver"))
    return AptPyVerIterator_new(self->pkg.TargetVer());
  else if(!strcmp(name, "currentver"))
    return AptPyVerIterator_new(self->pkg.CurrentVer());
  else
    {
      PyErr_SetString(PyExc_AttributeError, name);
      return NULL;
    }
}

static PyTypeObject AptPyPkgIterator_type=
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "AptPkgIterator",			// tp_name
  sizeof(AptPyPkgIterator),		// tp_basicsize
  0,					// tp_itemsize (undocumented)
  (destructor) AptPyPkgIterator_dealloc,// tp_dealloc
  0,					// tp_print
  (getattrfunc) AptPyPkgIterator_getattr,// tp_getattr
  0,					// tp_setattr
  (cmpfunc) AptPyPkgIterator_compare,	// tp_compare
  (reprfunc) AptPyPkgIterator_repr,	// tp_repr
  0,					// tp_as_number
  0,					// tp_as_mapping
  0,					// tp_hash
  0,					// tp_call
  0,					// tp_str
  0,					// tp_getattro
  0,					// tp_setattro
  0,					//  ???????
  0,					// tp_as_buffer
  Py_TPFLAGS_DEFAULT,			// tp_flags
  "A reference to a particular package in the package cache"
};

static PyObject *AptPyPkgIterator_new(pkgCache::PkgIterator pkg)
{
  AptPyPkgIterator *self=(AptPyPkgIterator *) PyObject_NEW(AptPyPkgIterator, &AptPyPkgIterator_type);

  if(!self)
    return NULL;

  self->pkg=pkg;
  return (PyObject *) self;
}

/*****************************************************************************
 * pkgCache::VerIterator wrapper
 *****************************************************************************/

struct AptPyVerIterator
{
  PyObject_HEAD;
  pkgCache::VerIterator ver;
};

static void AptPyVerIterator_dealloc(AptPyVerIterator *self)
{
  PyMem_DEL(self);
}

static int AptPyVerIterator_compare(AptPyVerIterator *a, AptPyVerIterator *b)
{
  return a->ver!=b->ver;
}

static PyObject *AptPyVerIterator_repr(AptPyVerIterator *self)
  // FIXME: should this be 'str' instead of 'repr'?
{
  if(self->ver.end())
    return PyString_FromString("<APT package iterator: END>");
  else
    return PyString_FromString((string("<APT package iterator: ")+self->ver.ParentPkg().Name()+" "+self->ver.VerStr()+">").c_str());
}

static PyObject *AptPyVerIterator_getattr(AptPyVerIterator *self, char *name)
{
  if(!strcmp(name, "end"))
    return PyInt_FromLong(self->ver.end());

  if(self->ver.end())
    {
      PyErr_SetString(PyExc_IndexError, "Attempt to index off the end of a version list");
      return NULL;
    }

  if(!strcmp(name, "next"))
    {
      pkgCache::VerIterator next=self->ver;
      next++;
      return AptPyVerIterator_new(next);
    }
  else if(!strcmp(name, "version") || !strcmp(name, "verstr"))
    return PyString_FromString(self->ver.VerStr());
  else if(!strcmp(name, "section"))
    return PyString_FromString(self->ver.Section());
  else if(!strcmp(name, "arch"))
    return PyString_FromString(self->ver.Arch());
  else if(!strcmp(name, "parentpkg"))
    return AptPyPkgIterator_new(self->ver.ParentPkg());
  else if(!strcmp(name, "prioritytype"))
    return PyString_FromString(self->ver.PriorityType());
  else if(!strcmp(name, "downloadable"))
    return PyInt_FromLong(self->ver.Downloadable());
  else if(!strcmp(name, "automatic"))
    return PyInt_FromLong(self->ver.Automatic());
  else if(!strcmp(name, "id"))
    return PyInt_FromLong(self->ver->ID);
  else if(!strcmp(name, "priority"))
    return PyInt_FromLong(self->ver->Priority);
  else
    {
      PyErr_SetString(PyExc_AttributeError, name);
      return NULL;
    }
}

static PyTypeObject AptPyVerIterator_type=
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "AptVerIterator",			// tp_name
  sizeof(AptPyVerIterator),		// tp_basicsize
  0,					// tp_itemsize (undocumented)
  (destructor) AptPyVerIterator_dealloc,// tp_dealloc
  0,					// tp_print
  (getattrfunc) AptPyVerIterator_getattr,// tp_getattr
  0,					// tp_setattr
  (cmpfunc) AptPyVerIterator_compare,	// tp_compare
  (reprfunc) AptPyVerIterator_repr,	// tp_repr
  0,					// tp_as_number
  0,					// tp_as_mapping
  0,					// tp_hash
  0,					// tp_call
  0,					// tp_str
  0,					// tp_getattro
  0,					// tp_setattro
  0,					//  ???????
  0,					// tp_as_buffer
  Py_TPFLAGS_DEFAULT,			// tp_flags
  "A reference to a particular version of a package in the package cache"
};

static PyObject *AptPyVerIterator_new(pkgCache::VerIterator ver)
{
  AptPyVerIterator *self=(AptPyVerIterator *) PyObject_NEW(AptPyVerIterator, &AptPyVerIterator_type);

  if(!self)
    return NULL;

  self->ver=ver;
  return (PyObject *) self;
}

/*****************************************************************************
 * pkgDepCache::State wrapper
 *****************************************************************************/

struct AptPyDepStateCache
{
  PyObject_HEAD;
  pkgDepCache::StateCache *state;
};

static PyObject *AptPyDepStateCache_new(pkgDepCache::StateCache *state);

static void AptPyDepStateCache_dealloc(AptPyDepStateCache *self)
{
  PyMem_DEL(self);
}

static PyObject *AptPyDepStateCache_getattr(AptPyDepStateCache *self, char *name)
{
  // FIXME: use a hash_map for this
  if(!strcmp(name, "candversion"))
    return PyString_FromString(self->state->CandVersion);
  else if(!strcmp(name, "curversion"))
    return PyString_FromString(self->state->CurVersion);
  else if(!strcmp(name, "status"))
    return PyInt_FromLong(self->state->Status);
  else if(!strcmp(name, "mode"))
    return PyInt_FromLong(self->state->Mode);
  else if(!strcmp(name, "depstate"))
    return PyInt_FromLong(self->state->DepState);
  else if(!strcmp(name, "flags"))
    return PyInt_FromLong(self->state->Flags);
  else if(!strcmp(name, "iflags"))
    return PyInt_FromLong(self->state->iFlags);
  else if(!strcmp(name, "newinstall"))
    return PyInt_FromLong(self->state->NewInstall());
  else if(!strcmp(name, "delete"))
    return PyInt_FromLong(self->state->Delete());
  else if(!strcmp(name, "keep"))
    return PyInt_FromLong(self->state->Keep());
  else if(!strcmp(name, "upgrade"))
    return PyInt_FromLong(self->state->Upgrade());
  else if(!strcmp(name, "upgradable"))
    return PyInt_FromLong(self->state->Upgradable());
  else if(!strcmp(name, "downgrade"))
    return PyInt_FromLong(self->state->Downgrade());
  else if(!strcmp(name, "held"))
    return PyInt_FromLong(self->state->Held());
  else if(!strcmp(name, "nowbroken"))
    return PyInt_FromLong(self->state->NowBroken());
  else if(!strcmp(name, "instbroken"))
    return PyInt_FromLong(self->state->InstBroken());
  else if(!strcmp(name, "install"))
    return PyInt_FromLong(self->state->Install());
  else
    {
      PyErr_SetString(PyExc_AttributeError, name);
      return NULL;
    }
}

static PyTypeObject AptPyDepStateCache_type=
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "pkgDepCache::StateCache",
  sizeof(AptPyDepStateCache),
  0,
  (destructor) AptPyDepStateCache_dealloc,
  0,
  (getattrfunc) AptPyDepStateCache_getattr,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  Py_TPFLAGS_DEFAULT,
  "Information about the state and planned action of a particular package"
};

static PyObject *AptPyDepStateCache_new(pkgDepCache::StateCache *state)
{
  AptPyDepStateCache *self=(AptPyDepStateCache *) PyObject_NEW(AptPyDepStateCache, &AptPyDepStateCache_type);

  if(self==NULL)
    return NULL;

  self->state=state;

  return (PyObject *) self;
}

/*****************************************************************************
 * pkgDepCache wrapper
 *
 *  FIXME: wrap MMap so the user can create his/her own depcaches (right now
 * you have to go through the rather limited pkgCacheFile class..)
 *****************************************************************************/

struct AptPyDepCache
{
  PyObject_HEAD;
  pkgDepCache *cache;
};

static PyObject *AptPyDepCache_new(pkgDepCache *cache);

static void AptPyDepCache_dealloc(AptPyDepCache *self)
{
  delete self->cache;
  PyMem_DEL(self);
}

// Mapping emulation:

static int AptPyDepCache_length(AptPyDepCache *self)
{
  return self->cache->Head().PackageCount;
}

static PyObject *AptPyDepCache_subscript(AptPyDepCache *self, PyObject *key)
{
  if(PyString_Check(key))
    {
      pkgCache::PkgIterator i=self->cache->FindPkg(PyString_AsString(key));
      if(i.end())
	return Py_None;
      else
	return AptPyPkgIterator_new(i);
    }
  else if(PyObject_Type(key)==(PyObject *) &AptPyPkgIterator_type)
    {
      pkgDepCache::StateCache &state=(*self->cache)[((AptPyPkgIterator *) key)->pkg];
      return AptPyDepStateCache_new(&state);
    }
  else
    {
      PyErr_SetObject(PyExc_TypeError, key);
      return NULL;
    }
}

static PyMappingMethods AptPyDepCache_as_mapping=
{
  (inquiry) AptPyDepCache_length,
  (binaryfunc) AptPyDepCache_subscript,
  0
};

static PyObject *AptPyDepCache_getattr(AptPyDepCache *self, char *name)
{
  if(!strcmp(name, "usrsize"))
    return PyLong_FromLong(self->cache->UsrSize());
  else if(!strcmp(name, "debsize"))
    return PyLong_FromLong(self->cache->DebSize());
  else if(!strcmp(name, "delcount"))
    return PyLong_FromLong(self->cache->DelCount());
  else if(!strcmp(name, "keepcount"))
    return PyLong_FromLong(self->cache->KeepCount());
  else if(!strcmp(name, "instcount"))
    return PyLong_FromLong(self->cache->InstCount());
  else if(!strcmp(name, "brokencount"))
    return PyLong_FromLong(self->cache->BrokenCount());
  else if(!strcmp(name, "badcount"))
    return PyLong_FromLong(self->cache->BadCount());
  else if(!strcmp(name, "packages"))
    return AptPyPkgIterator_new(self->cache->PkgBegin());
  else
    {
      PyErr_SetString(PyExc_AttributeError, name);
      return NULL;
    }
}

static PyTypeObject AptPyDepCache_type=
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "pkgDepCache::StateCache",
  sizeof(AptPyDepCache),
  0,
  (destructor) AptPyDepCache_dealloc,
  0,
  (getattrfunc) AptPyDepCache_getattr,
  0,
  0,
  0,
  0,
  0,
  &AptPyDepCache_as_mapping,
  0,
  0,
  0,
  0,
  0,
  0,
  Py_TPFLAGS_DEFAULT,
  "Cache of the dependency states of packages and what is to be done with them"
};

/*****************************************************************************
 * pkgCacheFile wrapper
 *****************************************************************************/

struct AptPyCacheFile
{
  PyObject_HEAD;
  pkgDepCache *cache;
  // cache files look a lot like depcaches..used to allow inheritence of some
  // stuff from pkgDepCache.
  pkgCacheFile *cachefile;
};

static PyObject *AptPyCacheFile_new(int withlock);

static void AptPyCacheFile_dealloc(AptPyCacheFile *self)
{
  delete self->cachefile;
  // Releases the lock, of course..
  PyMem_DEL(self);
}

static PyTypeObject AptPyCacheFile_type=
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "pkgCacheFile",
  sizeof(AptPyCacheFile),
  0,
  (destructor) AptPyCacheFile_dealloc,
  0,
  (getattrfunc) AptPyDepCache_getattr,
  0,
  0,
  0,
  0,
  0,
  &AptPyDepCache_as_mapping,
  0,
  0,
  0,
  0,
  0,
  0,
  Py_TPFLAGS_DEFAULT,
  """A wrapper around the dependency cache which first creates several locks to
ensure exclusive access to the files"""
};

static PyObject *AptPyCacheFile_fromprogbar(PyObject *self, PyObject *args)
{
  AptPyOpProgress *progress;
  int with_lock=1;

  pkgCacheFile *cachefile;
  if(!PyArg_ParseTuple(args, "O!|i", &AptPyOpProgress_type, &progress, &with_lock))
    return NULL;

  cachefile=new pkgCacheFile;
  if(!cachefile->Open(*progress->progress, with_lock))
    {
      delete cachefile;
      PyErr_SetString(AptPyExc_AptError, "Unable to lock or open cache file");
      return NULL;
    }

  AptPyCacheFile *rval=(AptPyCacheFile *) PyObject_NEW(AptPyCacheFile, &AptPyCacheFile_type);

  if(rval==NULL)
    return NULL;
  rval->cachefile=cachefile;
  rval->cache=&**cachefile;
  return (PyObject *) rval;
}

/*****************************************************************************
 * Global module declarations and initialization routine
 *****************************************************************************/

static PyMethodDef AptMethods[]=
{
  {"stringtobool", AptPyStringToBool, METH_VARARGS},
  {"stringtonum", AptPyStringToNum, METH_VARARGS},
  {"OpProgress", AptPyOpProgress_generic, METH_VARARGS},
  {"OpTextProgress", AptPyOpProgress_text, METH_VARARGS},
  {"MakeOpProgress", AptPyOpProgress_fromobj, METH_VARARGS},
  {"CacheFile", AptPyCacheFile_fromprogbar, METH_VARARGS},
  {NULL, NULL}
};

extern "C" void initapt()
{
  PyObject *module, *dict;

  module=Py_InitModule("apt", AptMethods);
  dict=PyModule_GetDict(module);

  pkgInitialize(*_config);

  PyObject *config_obj=AptPyConfig_new(_config);
  PyObject *error_obj=AptPyError_new(_error);
  // Don't let them get deallocated:
  Py_INCREF(config_obj);
  Py_INCREF(error_obj);

  PyDict_SetItemString(dict, "config", config_obj);
  PyDict_SetItemString(dict, "errors", error_obj);

  AptPyExc_AptError=PyErr_NewException("apt.AptError", NULL, NULL);
  PyDict_SetItemString(dict, "AptError", AptPyExc_AptError);
}
